@@ -29,8 +29,8 @@ from collections import defaultdict
 from typing import TYPE_CHECKING, Optional
 
 from aiohttp import web
-from aiorpcx import NetAddress
 
+from electrum import util
 from electrum.util import log_exceptions, ignore_exceptions
 from electrum.plugin import BasePlugin, hook
 from electrum.logging import Logger
@@ -60,7 +60,7 @@ class PayServerPlugin(BasePlugin):
         # we use the first wallet loaded
         if self.server is not None:
             return
-        if self.config.get('offline'):
+        if self.config.NETWORK_OFFLINE:
             return
         self.server = PayServer(self.config, wallet)
         asyncio.run_coroutine_threadsafe(daemon.taskgroup.spawn(self.server.run()), daemon.asyncio_loop)
@@ -79,8 +79,7 @@ class PayServer(Logger, EventListener):
         assert self.has_www_dir(), self.WWW_DIR
         self.config = config
         self.wallet = wallet
-        url = self.config.get('payserver_address', 'localhost:8080')
-        self.addr = NetAddress.from_string(url)
+        self.port = self.config.PAYSERVER_PORT
         self.pending = defaultdict(asyncio.Event)
         self.register_callbacks()
 
@@ -91,15 +90,11 @@ class PayServer(Logger, EventListener):
 
     @property
     def base_url(self):
-        payserver = self.config.get('payserver_address', 'localhost:8080')
-        payserver = NetAddress.from_string(payserver)
-        use_ssl = bool(self.config.get('ssl_keyfile'))
-        protocol = 'https' if use_ssl else 'http'
-        return '%s://%s:%d'%(protocol, payserver.host, payserver.port)
+        return 'http://localhost:%d'%self.port
 
     @property
     def root(self):
-        return self.config.get('payserver_root', '/r')
+        return self.config.PAYSERVER_ROOT
 
     @event_listener
     async def on_event_request_status(self, wallet, key, status):
@@ -118,13 +113,13 @@ class PayServer(Logger, EventListener):
         # to minimise attack surface. note: "add_routes" call order matters (inner path goes first)
         app.add_routes([web.static(f"{self.root}/vendor", os.path.join(self.WWW_DIR, 'vendor'), follow_symlinks=True)])
         app.add_routes([web.static(self.root, self.WWW_DIR)])
-        if self.config.get('payserver_allow_create_invoice'):
+        if self.config.PAYSERVER_ALLOW_CREATE_INVOICE:
             app.add_routes([web.post('/api/create_invoice', self.create_request)])
         runner = web.AppRunner(app)
         await runner.setup()
-        site = web.TCPSite(runner, host=str(self.addr.host), port=self.addr.port, ssl_context=self.config.get_ssl_context())
+        site = web.TCPSite(runner, host='localhost', port=self.port)
         await site.start()
-        self.logger.info(f"now running and listening. addr={self.addr}")
+        self.logger.info(f"running and listening on port {self.port}")
 
     async def create_request(self, request):
         params = await request.post()
@@ -146,7 +141,7 @@ class PayServer(Logger, EventListener):
         return web.json_response(request)
 
     async def get_bip70_request(self, r):
-        from .paymentrequest import make_request
+        from electrum.paymentrequest import make_request
         key = r.match_info['key']
         request = self.wallet.get_request(key)
         if not request:
@@ -173,7 +168,7 @@ class PayServer(Logger, EventListener):
             return ws
         while True:
             try:
-                await asyncio.wait_for(self.pending[key].wait(), 1)
+                await util.wait_for2(self.pending[key].wait(), 1)
                 break
             except asyncio.TimeoutError:
                 # send data on the websocket, to keep it alive
